package checkers.nullness;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;
import javax.lang.model.util.ElementFilter;
import javax.lang.model.util.Elements;

import checkers.types.AnnotatedTypeMirror.AnnotatedDeclaredType;
import checkers.util.InternalUtils;
import checkers.util.TreeUtils;

import com.sun.source.tree.*;
import com.sun.source.util.TreePath;

/**
 * Utilities class for handling {@code Map.get()} invocations.
 */
/*package-scope*/ class MapGetHeauristics {

    private final ProcessingEnvironment env;
    private final Elements elements;
    private final NullnessAnnotatedTypeFactory factory;

    private final ExecutableElement mapGet;
    private final ExecutableElement mapKeySet;
    private final ExecutableElement mapContains;

    public MapGetHeauristics(ProcessingEnvironment env,
            NullnessAnnotatedTypeFactory factory) {
        this.env = env;
        this.elements = env.getElementUtils();
        this.factory = factory;

        mapGet = getMethod("java.util.Map", "get", 1);
        mapKeySet = getMethod("java.util.Map", "keySet", 0);
        mapContains = getMethod("java.util.Map", "containsKey", 1);
    }

    /**
     * @return true if the tree is a method invocation to a method overriding
     *        {@code Map.get(K)}
     */
    public boolean isMapGetInvocation(Tree tree) {
        return isMethod(tree, mapGet);
    }

    /**
     * Checks whether the key passed to {@code Map.get(K key)} is known
     * to be in the map.
     * 
     * TODO: Document when this method returns true
     */
    public boolean isSuppressable(MethodInvocationTree tree) {
        Element elt = getSite(tree);
        return (elt instanceof VariableElement
                && tree.getArguments().get(0) instanceof IdentifierTree
                && isKeyInMap((IdentifierTree)tree.getArguments().get(0), (VariableElement)elt));
    }

    private boolean isKeyInMap(IdentifierTree tree, VariableElement map) {
        TreePath path = factory.getPath(tree);
        Element key = TreeUtils.elementFromUse(tree);

        // Check for contains
        IfTree ifTree = (IfTree) TreeUtils.enclosingOfKind(path, Tree.Kind.IF);
        if (ifTree != null
                && TreeUtils.skipParens(ifTree.getCondition()) instanceof MethodInvocationTree) {
            MethodInvocationTree invok = (MethodInvocationTree)TreeUtils.skipParens(ifTree.getCondition());
            if (isMethod(invok, mapContains)) {
                Element containsArgument = InternalUtils.symbol(invok.getArguments().get(0));
                return key.equals(containsArgument) && map.equals(getSite(invok));
            }
        }
        
        // Check for enhanced for loop
        EnhancedForLoopTree forLoop = (EnhancedForLoopTree) TreeUtils.enclosingOfKind(path, Tree.Kind.ENHANCED_FOR_LOOP);
        if (forLoop != null
                && forLoop.getExpression() instanceof MethodInvocationTree) {
            MethodInvocationTree iterable = (MethodInvocationTree)forLoop.getExpression();
            return (key.equals(TreeUtils.elementFromDeclaration(forLoop.getVariable()))
                    && isMethod(iterable, mapKeySet)
                    && map.equals(getSite(iterable)));
        }
        return false;
    }

    private Element getSite(MethodInvocationTree tree) {
        AnnotatedDeclaredType type =
            (AnnotatedDeclaredType)factory.getReceiver(tree);
        return type.getElement();
    }

    private boolean isMethod(Tree tree, ExecutableElement method) {
        if (!(tree instanceof MethodInvocationTree))
            return false;
        MethodInvocationTree methInvok = (MethodInvocationTree)tree;
        ExecutableElement invoked = TreeUtils.elementFromUse(methInvok);
        return isMethod(invoked, method);
    }

    private boolean isMethod(ExecutableElement questioned, ExecutableElement method) {
        return (questioned.equals(method)
                || env.getElementUtils().overrides(questioned, method,
                        (TypeElement)questioned.getEnclosingElement()));
    }

    private ExecutableElement getMethod(String typeName, String methodName, int params) {
        TypeElement mapElt = env.getElementUtils().getTypeElement(typeName);
        for (ExecutableElement exec : ElementFilter.methodsIn(mapElt.getEnclosedElements())) {
            if (exec.getSimpleName().contentEquals(methodName)
                    && exec.getParameters().size() == params)
                return exec;
        }
        throw new RuntimeException("Shouldn't be here!");
    }
}
